/*
 * RFCNB (RFC1001/RFC1002) NEtBIOS implementation
 * 
 * Version 1.0
 * RFCNB IO Routines ...
 * 
 * Copyright (C) Richard Sharpe 1996
 * 
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"


int RFCNB_Timeout = 0;       /* Timeout in seconds ... */

static void rfcnb_alarm (int sig)
{
  fprintf (stderr, "IO Timed out ...\n");
}

/*
 * Set timeout value and setup signal handling
 */
int RFCNB_Set_Timeout (int seconds)
{
  RFCNB_Timeout = seconds;

  if (RFCNB_Timeout > 0)
  {
    struct sigaction inact, outact;

    inact.sa_handler = (void (*)()) rfcnb_alarm;
    sigemptyset (&inact.sa_mask);
    inact.sa_flags = 0;         /* Don't restart */

    if (sigaction (SIGALRM, &inact, &outact) < 0)
       return (-1);
  }
  return (0);
}

/*
 * Discard the rest of an incoming packet as we do not have space for it
 * in the buffer we allocated or were passed ...
 */
static int Discard_Rest (struct RFCNB_Con *con, int len)
{
  char temp[100];
  int  rest = len;
  int  this_read, bytes_read;

  SMB_debug (2, "Discard_Rest: length %d\n", len);

  while (rest > 0)
  {
    this_read  = (rest > sizeof (temp) ? sizeof (temp) : rest);
    bytes_read = SOCKET_READ (con->fd, temp, this_read);

    if (bytes_read <= 0)   /* Error so return */
    {    
      if (bytes_read < 0)
           RFCNB_errno = RFCNBE_BadRead;
      else RFCNB_errno = RFCNBE_ConGone;
      RFCNB_saved_errno = errno;
      return (-1);
    }
    rest -= bytes_read;
  }
  return (0);
}


/*
 * Send an RFCNB packet to the connection.
 * 
 * We just send each of the blocks linked together ... 
 * 
 * If we can, try to send it as one iovec ... 
 * 
 */
int RFCNB_Put_Pkt (struct RFCNB_Con *con, struct RFCNB_Pkt *pkt, int len)
{
  struct RFCNB_Pkt *pkt_ptr;
  struct iovec      io_list[10];   /* We should never need more */
  int    i, len_sent, tot_sent, this_len;
  char  *this_data;

  /* Try to send the data ... We only send as many bytes as len claims.
   * We should try to stuff it into an IOVEC and send as one write
   */
  pkt_ptr  = pkt;
  len_sent = tot_sent = 0;         /* Nothing sent so far */
  i = 0;

  while (pkt_ptr && i < DIM(io_list))
  {
    this_len  = pkt_ptr->len;
    this_data = pkt_ptr->data;
    if (tot_sent + this_len > len)
       this_len = len - tot_sent;  /* Adjust so we don't send too much */

    /* Now plug into the iovec ...
     */
    io_list[i].iov_len  = this_len;
    io_list[i].iov_base = this_data;
    i++;

    tot_sent += this_len;
    if (tot_sent == len)
       break;                      /* Let's not send too much */

    pkt_ptr = pkt_ptr->next;
  }

  SMB_debug (2, "frags = %d, tot_sent = %d\n", i, tot_sent);

  /* Set up an alarm if timeouts are set ...
   */
  if (RFCNB_Timeout > 0)
     alarm (RFCNB_Timeout);

  len_sent = SOCKET_WRITEV (con->fd, io_list, i);
  if (len_sent < 0)
  {
    con->errno = errno;
    if (errno == EINTR || errno == ETIMEDOUT)
         RFCNB_errno = RFCNBE_Timeout;
    else RFCNB_errno = RFCNBE_BadWrite;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  if (len_sent < tot_sent)         /* Less than we wanted */
  {                  
    if (errno == EINTR || errno == ETIMEDOUT)
         RFCNB_errno = RFCNBE_Timeout;
    else RFCNB_errno = RFCNBE_BadWrite;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  if (RFCNB_Timeout > 0)
     alarm (0);                    /* Reset that sucker */

  SMB_debug (2, "tx-len = %d ...\n", len_sent);
  RFCNB_Print_Pkt (stderr, "sent", pkt, len_sent);
  return (len_sent);
}

/*
 * Read an RFCNB packet off the connection.
 * 
 * We read the first 4 bytes, that tells us the length, then read the
 * rest. We should implement a timeout, but we don't just yet 
 * 
 */
int RFCNB_Get_Pkt (struct RFCNB_Con *con, struct RFCNB_Pkt *pkt, int len)
{
  struct RFCNB_Pkt *pkt_frag;
  struct RFCNB_Hdr  hdr;  /* !! was "char hdr[RFCNB_Pkt_Hdr_Len]" */

  int    read_len, pkt_len;
  int    more, this_time, offset;
  int    frag_len, this_len;
  BOOL   seen_keep_alive = TRUE;

  /* Read that header straight into the buffer
   */
  if (len < sizeof(hdr))
  {  
    SMB_debug (1, "short packet: %s\n", strerror(errno));
    RFCNB_errno = RFCNBE_BadParam;
    return (-1);
  }

  /* We discard keep alives here ...
   */
  if (RFCNB_Timeout > 0)
    alarm (RFCNB_Timeout);

  while (seen_keep_alive)
  {
    read_len = SOCKET_READ (con->fd, (void*)&hdr, sizeof(hdr));
    if (read_len < 0)
    {
      SMB_debug (1, "SOCKET_READ: %s\n", strerror(errno));

      if (errno == EINTR || errno == ETIMEDOUT)
           RFCNB_errno = RFCNBE_Timeout;
      else RFCNB_errno = RFCNBE_BadRead;
      RFCNB_saved_errno = errno;
      return (-1);
    }

    /* Now we check out what we got
     */
    if (read_len == 0)   /* Connection closed? */
    {
      SMB_debug (1, "read 0-bytes: %s\n", strerror(errno));

      if (errno == EINTR || errno == ETIMEDOUT)
           RFCNB_errno = RFCNBE_Timeout;
      else RFCNB_errno = RFCNBE_ConGone;
      RFCNB_saved_errno = errno;
      return (-1);
    }

    if (RFCNB_Pkt_Type (&hdr) == RFCNB_SESSION_KEEP_ALIVE)
         SMB_debug (2, "RFCNB KEEP ALIVE received\n");
    else seen_keep_alive = FALSE;
  }

  /* What if we got less than or equal to a hdr size in bytes?
   */
  if (read_len < sizeof(hdr))
  {
    /* Now we need to copy the hdr portion we got into the supplied packet
     */
    memcpy (pkt->data, &hdr, read_len);

    RFCNB_Print_Pkt (stderr, "rcvd", pkt, read_len);

    return (read_len);
  }

  /* Now, if we got at least a hdr size, alloc space for rest,
   * if we need it
   */
  pkt_len = RFCNB_Pkt_Len (&hdr);

  SMB_debug (2, "Reading Pkt: length = %d\n", pkt_len);

  /* Now copy in the header
   */
  memcpy (pkt->data, &hdr, sizeof(hdr));

  /* Get the rest of the packet ... first figure out how big our buf is?
   * And make sure that we handle the fragments properly.
   */

  if (len < pkt_len)            /* Only get as much as we have space for */
       more = len - RFCNB_Pkt_Hdr_Len;
  else more = pkt_len;

  this_time = 0;

  /* We read for each fragment ...
   */
  if (pkt->len == read_len)
  {                             /* If this frag was exact size */
    pkt_frag = pkt->next;       /* Stick next lot in next frag */
    offset = 0;                 /* then we start at 0 in next  */
  }
  else
  {
    pkt_frag = pkt;             /* Otherwise use rest of this frag */
    offset = RFCNB_Pkt_Hdr_Len; /* Otherwise skip the header */
  }

  frag_len = pkt_frag->len;

  if (more <= frag_len)         /* If len left to get less than frag space */
       this_len = more;         /* Get the rest ... */
  else this_len = frag_len - offset;

  while (more > 0)
  {
    this_time = SOCKET_READ (con->fd, (pkt_frag->data) + offset, this_len);
    if (this_time <= 0)
    {
      if (errno == EINTR || errno == ETIMEDOUT)
      {
        RFCNB_errno = RFCNB_Timeout;
        RFCNB_saved_errno = errno;
      }
      else
      {
        if (this_time < 0)
             RFCNB_errno = RFCNBE_BadRead;
        else RFCNB_errno = RFCNBE_ConGone;
        RFCNB_saved_errno = errno;
      }
      return (-1);
    }

    SMB_debug (2, "frag_len = %d, this_time = %d, this_len = %d, more = %d\n",
               frag_len, this_time, this_len, more);

    read_len = read_len + this_time; /* How much have we read ... */

    if (!pkt_frag->next)             /* stop if no next part */
       break;

    pkt_frag = pkt_frag->next;
    this_len = pkt_frag->len;
    offset   = 0;
    more    -= this_time;
  }

  SMB_debug (2, "pkt_len = %d, read_len = %d\n", pkt_len, read_len);
  RFCNB_Print_Pkt (stderr, "rcvd", pkt, read_len + sizeof(hdr));

  if (read_len < (pkt_len + sizeof(hdr))) /* Discard the rest */
     return Discard_Rest (con, pkt_len + sizeof(hdr) - read_len);

  /* !!to-do: make this portable
   */
  if (RFCNB_Timeout > 0)
    alarm (0);                  /* Reset that sucker */

  return (read_len + sizeof(hdr));
}
