/*
 * RFCNB (RFC1001/RFC1002) NetBIOS implementation
 * 
 * Version 1.0
 * Session Routines ...
 * 
 * Copyright (C) Richard Sharpe 1996
 * 
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"

int RFCNB_errno       = 0;
int RFCNB_saved_errno = 0;

int RFCNB_Stats [RFCNB_MAX_STATS];

void (*Prot_Print_Routine)() = NULL; /* Pointer to print routine */

/*
 * Text strings for the error responses
 */
static char *RFCNB_Error_Strings[] = {

  "RFCNBE_OK: Routine completed successfully.",
  "RFCNBE_NoSpace: No space available for a malloc call.",
  "RFCNBE_BadName: NetBIOS name could not be translated to IP address.",
  "RFCNBE_BadRead: Read system call returned an error. Check errno.",
  "RFCNBE_BadWrite: Write system call returned an error. Check errno.",
  "RFCNBE_ProtErr: A protocol error has occurred.",
  "RFCNBE_ConGone: Connection dropped during a read or write system call.",
  "RFCNBE_BadHandle: Bad connection handle passed.",
  "RFCNBE_BadSocket: Problems creating socket.",
  "RFCNBE_ConnectFailed: Connection failed. See errno.",
  "RFCNBE_CallRejNLOCN: Call rejected. Not listening on called name.",
  "RFCNBE_CallRejNLFCN: Call rejected. Not listening for called name.",
  "RFCNBE_CallRejCNNP: Call rejected. Called name not present.",
  "RFCNBE_CallRejInfRes: Call rejected. Name present, but insufficient resources.",
  "RFCNBE_CallRejUnSpec: Call rejected. Unspecified error.",
  "RFCNBE_BadParam: Bad parameters passed to a routine.",
  "RFCNBE_Timeout: IO Operation timed out ..."
};

/*
 * Set up a session with a remote name. We are passed 'called_name' as a
 * string which we convert to a NetBIOS name, i.e. space terminated, up to
 * 16 characters only if we need to. If 'called_addr' is not empty, then
 * we use it to connect to the remote end, but put in 'called_name'.
 * 'called_addr' can be a DNS based name, or a TCP/IP address.
 */
struct RFCNB_Con *RFCNB_Call (const char *called_name,
                              const char *calling_name,
                              const char *called_addr,
                              unsigned short port)
{
  struct RFCNB_Con     *con;
  struct redirect_addr *redir_addr;
  struct in_addr        dest_IP;
  int    client;
  BOOL   redirect;
  const  char *service_addr;

  /* Look up the "netbios-session" port in /etc/services ...
   */
  if (port == 0)
  {
    struct servent *se = getservbyname ("udp", "netbios-ssn");
    if (se)
         port = ntohs (se->s_port);
    else port = RFCNB_DEFAULT_PORT;
  }

  /* Create a connection structure first
   */
  con = calloc (sizeof(*con), 1);
  if (!con)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    return (NULL);
  }

  /* Resolve that name into an IP address
   */
  service_addr = called_name;
  if (called_addr && *called_addr)
     service_addr = called_addr;

  if (RFCNB_Name_To_IP (service_addr, &dest_IP) < 0)
  {
    RFCNB_saved_errno = errno;
    return (NULL);
  }

  /* Now try to connect to the remote
   */
  redirect = TRUE;              /* Fudge this one so we go once through */

  while (redirect)
  {                             /* Connect and get session info etc */
    redirect = FALSE;           /* Assume all OK */

    /* Build the redirect info. First one is first addr called
     * And tack it onto the list of addresses we called
     */
    redir_addr = calloc (sizeof(*redir_addr), 1);
    if (!redir_addr)
    {
      RFCNB_errno = RFCNBE_NoSpace;
      RFCNB_saved_errno = errno;
      return (NULL);
    }

    memcpy (&redir_addr->ip_addr, &dest_IP, sizeof(dest_IP));
    redir_addr->port = port;

    if (!con->redirect_list)    /* Stick on head */
    {
      con->redirect_list = redir_addr;
      con->last_addr     = redir_addr;
    }
    else
    {   
      con->last_addr->next = redir_addr;
      con->last_addr       = redir_addr;
    }

    /* Now, make that connection
     */
    client = RFCNB_IP_Connect (dest_IP, port);
    if (client < 0)
       return (NULL);

    con->fd = client;

    /* Now send and handle the RFCNB session request.
     * If we get a redirect, we will comeback with redirect true
     * and a new IP address in dest_IP
     */
    if (RFCNB_Session_Req (con, called_name, calling_name,
                           &redirect, &dest_IP, &port) < 0)
    {
      RFCNB_saved_errno = errno;
      return (NULL);
    }

    if (redirect)
    {
      /* We have to close the connection, and then try again
       */
      con->redirects++;
      RFCNB_Close (con->fd); 
    }
  }
  return (con);
}

/*
 * We send a packet to the other end ... for the moment, we treat the
 * data as a series of pointers to blocks of data ... we should check the
 * length ...
 */
int RFCNB_Send (struct RFCNB_Con *con, struct RFCNB_Pkt *udata, int length)
{
  struct RFCNB_Pkt *pkt;
  char   *hdr;
  int     len;

  /* Plug in the header and send the data
   */
  pkt = RFCNB_Alloc_Pkt (RFCNB_Pkt_Hdr_Len);
  if (!pkt)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  pkt->next = udata;     /* The user data we want to send */
  hdr = pkt->data;

  /* Following crap is for portability across multiple UNIX machines
   */
  *(hdr + RFCNB_Pkt_Type_Offset) = RFCNB_SESSION_MESSAGE;
  RFCNB_Put_Pkt_Len (hdr, length);

  SMB_debug (2, "Sending packet: ");

  len = RFCNB_Put_Pkt (con, pkt, length + RFCNB_Pkt_Hdr_Len);
  if (len < 0)
     return (-1);

  /* Now we have sent that lot, let's get rid of the RFCNB Header and return
   */
  pkt->next = NULL;
  RFCNB_Free_Pkt (pkt);
  return (len);
}

/*
 * We pick up a message from the internet ... We have to worry about
 * non-message packets ...
 */
int RFCNB_Recv (struct RFCNB_Con *con, struct RFCNB_Pkt *data, int length)
{
  struct RFCNB_Pkt *pkt;
  int    rc;

  if (!con)
  {
    RFCNB_errno = RFCNBE_BadHandle;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  /* Now get a packet from below. We allocate a header first
   * Plug in the header and send the data
   */
  pkt = RFCNB_Alloc_Pkt (RFCNB_Pkt_Hdr_Len);
  if (!pkt)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    return (-1); 
  }

  pkt->next = data;           /* Plug in the data portion */
  rc = RFCNB_Get_Pkt (con, pkt, length + RFCNB_Pkt_Hdr_Len);
  if (rc < 0)
  {
    SMB_debug (1, "Bad packet return in RFCNB_Recv... \n");
    return (-1);
  }

  /* We should check that we go a message and not a keep alive
   */
  pkt->next = NULL;
  RFCNB_Free_Pkt (pkt);
  return (rc);
}

/*
 * We just disconnect from the other end, as there is nothing in the RFCNB
 * protocol that specifies any exchange as far as I can see
 */
int RFCNB_Hangup (struct RFCNB_Con *con)
{
  if (con)
  {
    RFCNB_Close (con->fd); /* Could this fail? */
    free (con);
  }
  return (0);
}

/*
 * Set TCP_NODELAY on the socket
 */
int RFCNB_Set_Sock_NoDelay (const struct RFCNB_Con *con, BOOL yn)
{
  int on = yn;
  return setsockopt (con->fd, IPPROTO_TCP, TCP_NODELAY,
                     (char*)&on, sizeof(on));
}

/*
 * Listen for a connection on a port???, when
 * the connection comes in, we return with the connection
 */
struct RFCNB_Con *RFCNB_Listen (unsigned short port)
{
  ARGSUSED (port);
  return (NULL);
}

/*
 * Pick up the last error response as a string, hmmm, this routine should
 * have been different ...
 */
void RFCNB_Get_Error (char *buffer, int buf_len)
{
  char *p   = buffer;
  char *err = RFCNB_Error_Strings [RFCNB_errno];

#ifdef HAVE_SNPRINTF
  p += snprintf (p, buf_len, "%s", err);
  if (p > buffer && RFCNB_saved_errno > 0)
     sprintf (p, buf_len - (p - buffer),
              "\n\terrno: %s", strerror(RFCNB_saved_errno));
#else
  p += sprintf (p, "%s", err);
  if (RFCNB_saved_errno > 0)
     sprintf (p, "\n\terrno: %s", strerror(RFCNB_saved_errno));
  ARGSUSED (buf_len);
#endif
}

/*
 * Pick up the last error response and returns as a code
 */
int RFCNB_Get_Last_Error (void)
{
  return (RFCNB_errno);
}

/*
 * Pick up saved errno as well
 */
int RFCNB_Get_Last_Errno (void)
{
  return (RFCNB_saved_errno);
}

/*
 * Pick up the last error response and return in string ...
 */
int RFCNB_Get_Error_Msg (int code, char *msg_buf, int len)
{
  strncpy (msg_buf, RFCNB_Error_Strings[abs(code)], len-1);
  msg_buf[len-1] = '\0';
  return (code);
}

/*
 * Register a higher level protocol print routine
 */
void RFCNB_Register_Print_Routine (void (*fn)())
{
  Prot_Print_Routine = fn;
}
